@@ -1,5 +1,6 @@
+import asyncio
 import json
-from typing import Any, List, Tuple, Union, cast
+from typing import Any, Callable, List, Tuple, Union, cast
 
 from eth_account import Account
 from eth_account.datastructures import SignedMessage, SignedTransaction
@@ -9,12 +10,12 @@ from eth_typing import Address
 from eth_typing.encoding import HexStr
 from hexbytes import HexBytes
 from web3 import Web3
-from web3.contract import Contract, ContractFunction
+from web3.contract.contract import Contract, ContractFunction
 from web3.gas_strategies import rpc
 from web3.types import (
     BlockData,
     BlockIdentifier,
-    CallOverrideParams,
+    CallOverride,
     Middleware,
     Nonce,
     TxData,
@@ -22,8 +23,9 @@ from web3.types import (
     TxReceipt,
     Wei,
 )
+from websockets.client import connect
 
-from web3client.exceptions import TransactionTooExpensive
+from web3client.exceptions import TransactionTooExpensive, Web3ClientException
 
 
 class BaseClient:
@@ -113,7 +115,7 @@ class BaseClient:
 
     def set_contract(self, contract_address: Address, abi: dict[str, Any]) -> None:
         self.contract_address: Address = cast(
-            Address, Web3.toChecksumAddress(contract_address)
+            Address, Web3.to_checksum_address(contract_address)
         )
         self.abi: dict[str, Any] = abi
         self.contract = self.getContract(contract_address, self.w3, abi=abi)
@@ -159,7 +161,7 @@ class BaseClient:
         if self.tx_type == 1:
             self.w3.eth.set_gas_price_strategy(rpc.rpc_gas_price_strategy)
             tx["gasPrice"] = self.w3.eth.generate_gas_price()
-            gas_fee_in_gwei = float(Web3.fromWei(tx["gasPrice"], "gwei"))
+            gas_fee_in_gwei = float(Web3.from_wei(tx["gasPrice"], "gwei"))
 
         # Post EIP-1599, we have both the miner's tip and the max fee.
         elif self.tx_type == 2:
@@ -169,13 +171,13 @@ class BaseClient:
             max_priority_fee_in_gwei = (
                 max_priority_fee_in_gwei or self.max_priority_fee_in_gwei
             )
-            tx["maxPriorityFeePerGas"] = Web3.toWei(max_priority_fee_in_gwei, "gwei")
+            tx["maxPriorityFeePerGas"] = Web3.to_wei(max_priority_fee_in_gwei, "gwei")
 
             # The max fee is estimated from the miner tip & block base fee
             (maxFeePerGasInGwei, gas_fee_in_gwei) = self.estimate_max_fee_in_gwei(
                 max_priority_fee_in_gwei
             )
-            tx["maxFeePerGas"] = Web3.toWei(maxFeePerGasInGwei, "gwei")
+            tx["maxFeePerGas"] = Web3.to_wei(maxFeePerGasInGwei, "gwei")
 
         # Raise an exception if the fee is too high
         self.raise_if_gas_fee_too_high(gas_fee_in_gwei)
@@ -187,7 +189,7 @@ class BaseClient:
         # For more details, see docs of ContractFunction.transact in
         # https://web3py.readthedocs.io/en/stable/contracts.html
         if gas_limit is not None:
-            tx["gas"] = gas_limit  # type: ignore
+            tx["gas"] = gas_limit
 
         return tx
 
@@ -203,7 +205,7 @@ class BaseClient:
         Build a transaction involving a transfer of value (in ETH) to an address,
         where the value is expressed in the blockchain token (e.g. ETH or AVAX).
         """
-        value_in_wei = self.w3.toWei(value_in_eth, "ether")
+        value_in_wei = self.w3.to_wei(value_in_eth, "ether")
         return self.build_tx_with_value_in_wei(
             to, value_in_wei, nonce, gas_limit, max_priority_fee_in_gwei
         )
@@ -228,7 +230,7 @@ class BaseClient:
         extra_params: TxParams = {
             "to": to,
             "value": value_in_wei,
-            "gas": self.estimate_gas_for_transfer(to, value_in_wei),  # type: ignore
+            "gas": self.estimate_gas_for_transfer(to, value_in_wei),
         }
         return tx | extra_params
 
@@ -253,7 +255,7 @@ class BaseClient:
         )
         if value_in_wei:
             base_tx["value"] = value_in_wei
-        return contract_function.buildTransaction(base_tx)
+        return contract_function.build_transaction(base_tx)
 
     ####################
     # Sign & send Tx
@@ -271,7 +273,7 @@ class BaseClient:
         Send a signed transaction and return the tx hash
         """
         tx_hash = self.w3.eth.send_raw_transaction(signed_tx.rawTransaction)
-        return self.w3.toHex(tx_hash)
+        return self.w3.to_hex(tx_hash)
 
     def sign_and_send_tx(self, tx: TxParams) -> HexStr:
         """
@@ -347,23 +349,85 @@ class BaseClient:
         return signer_address == self.user_address
 
     ####################
-    # Watch
+    # Scan
     ####################
 
-    # def watch(
-    #         eventName: str,
-    #         argument_filters: Optional[Dict[str, Any]] = None,
-    #         fromBlock: Optional[BlockIdentifier] = None,
-    #         toBlock: BlockIdentifier = "latest",
-    #         address: Optional[ChecksumAddress] = None,
-    #         topics: Optional[Sequence[Any]] = None) -> None:
-    #     """
-    #     Watch for a certain event
-    #     """
-    #     method_to_call = getattr(foo, 'bar')
-    #     result = method_to_call()
-    #     event = client.contract.events.StartGame()
-    #     filter = event.createFilter(fromBlock='latest' ...)
+    def subscribe_pending_txs(
+        self,
+        on_tx: Callable[[str], None],
+        on_subscribe: Callable[[str], None] = None,
+        once: bool = False,
+        subscription_type: str = "newPendingTransactions",
+    ) -> None:
+        """Asynchronously scan pending transactions with eth_subscribe and call the
+        given callback when one is found.
+
+        Details:
+         - The callback is called with the transaction hash as argument.
+         - The subscription type is 'newPendingTransactions' by default.
+         - Provid once=True to stop the subscription after the first transaction is found.
+
+        Caveats:
+         - The client must be configured with a websocket RPC endpoint (ws:// or wss://)
+         - Not all chains support the 'newPendingTransactions' subscription type.
+         - Not all chains have a mempool, e.g. Arbitrum. For these chains, the function will
+           just hang.
+         - Some chains require a validator node with staked L1 funds to be able to
+           access to pending transactions in the mempool (e.g. Avalanche).
+         - If you use Alchemy, you might want to use 'alchemy_newPendingTransactions'
+           (https://docs.alchemy.com/reference/newpendingtransactions)
+        """
+        # Raise if not a websocket uri
+        rpc_url = self.node_uri
+        if not rpc_url.startswith("ws"):
+            raise ValueError(
+                "Websocket RPC needed to subscribe to pending transactions"
+            )
+
+        async def get_event() -> None:
+            # Connect to websocket
+            async with connect(self.node_uri) as ws:
+                # Subscribe to pending transactions
+                await ws.send(
+                    '{"jsonrpc": "2.0", "id": 1, "method": "eth_subscribe", "params": ["'
+                    + subscription_type
+                    + '"]}'
+                )
+                subscription_response = await ws.recv()
+                try:
+                    subscription_id = json.loads(subscription_response)["result"]
+                except Exception as e:
+                    raise Web3ClientException(
+                        f"Failed to subscribe to pending transactions: {e}"
+                    )
+                # Call on_subscribe callback
+                if on_subscribe:
+                    on_subscribe(json.loads(subscription_response))
+                while True:
+                    # Wait for new transactions
+                    tx_response = await asyncio.wait_for(ws.recv(), timeout=15)
+                    try:
+                        tx_response_json = json.loads(tx_response)
+                        tx_subscription = tx_response_json["params"]["subscription"]
+                        tx_hash = tx_response_json["params"]["result"]
+                    except Exception as e:
+                        raise Web3ClientException(
+                            f"Failed to parse pending transaction: {e}"
+                        )
+                    # Call on_tx callback
+                    if tx_subscription == subscription_id:
+                        on_tx(tx_hash)
+                        if once:
+                            raise StopAsyncIteration
+
+        # Run loop asynchronously
+        loop = asyncio.get_event_loop()
+        while True:
+            try:
+                loop.run_until_complete(get_event())
+            except StopAsyncIteration:
+                loop.close()
+                break
 
     ####################
     # Gas
@@ -386,7 +450,7 @@ class BaseClient:
         """
         latest_block = self.w3.eth.get_block("latest")
         base_fee_in_wei = latest_block["baseFeePerGas"]
-        base_fee_in_gwei = float(Web3.fromWei(base_fee_in_wei, "gwei"))
+        base_fee_in_gwei = float(Web3.from_wei(base_fee_in_wei, "gwei"))
         return (2 * base_fee_in_gwei + max_priority_fee_in_gwei, base_fee_in_gwei)
 
     def estimate_gas_price_in_gwei(self) -> float:
@@ -397,7 +461,7 @@ class BaseClient:
         Docs: https://web3py.readthedocs.io/en/stable/gas_price.html
         """
         self.w3.eth.set_gas_price_strategy(rpc.rpc_gas_price_strategy)
-        return float(Web3.fromWei(self.w3.eth.generate_gas_price(), "gwei"))
+        return float(Web3.from_wei(self.w3.eth.generate_gas_price(), "gwei"))
 
     def raise_if_gas_fee_too_high(self, gas_fee_in_gwei: float) -> None:
         """
@@ -449,10 +513,10 @@ class BaseClient:
         )
 
     def get_balance_in_wei(self, address: Address = None) -> Wei:
-        return self.w3.eth.get_balance(Web3.toChecksumAddress(address))
+        return self.w3.eth.get_balance(Web3.to_checksum_address(address))
 
     def get_balance_in_eth(self, address: Address = None) -> float:
-        return float(Web3.fromWei(self.get_balance_in_wei(address), "ether"))
+        return float(Web3.from_wei(self.get_balance_in_wei(address), "ether"))
 
     ####################
     # Contract
@@ -463,7 +527,7 @@ class BaseClient:
         function: ContractFunction,
         tx: TxParams = None,
         block_identifier: BlockIdentifier = "latest",
-        state_override: CallOverrideParams = None,
+        state_override: CallOverride = None,
     ) -> Any:
         """
         Execute a contract function call using the eth_call interface.
@@ -513,7 +577,7 @@ class BaseClient:
         Load the smart contract, required before running
         build_contract_tx().
         """
-        checksum = Web3.toChecksumAddress(address)
+        checksum = Web3.to_checksum_address(address)
         if abi_file:
             abi = BaseClient.get_contract_abi_from_file(abi_file)
         return provider.eth.contract(address=checksum, abi=abi)
@@ -545,5 +609,7 @@ class BaseClient:
         was spent in gas to process the transaction
         """
         return float(
-            Web3.fromWei(txReceipt["effectiveGasPrice"] * txReceipt["gasUsed"], "ether")
+            Web3.from_wei(
+                txReceipt["effectiveGasPrice"] * txReceipt["gasUsed"], "ether"
+            )
         )
